from typing import Optional

import isotp
from isotp.address import AddressingMode
from pydantic import Base64Bytes, BaseModel


class CanMessage(BaseModel):
    """
    Internal CAN message type used for gRPC transmission.
    """

    timestamp: float
    arbitration_id: int
    is_extended_id: bool
    is_remote_frame: bool
    is_error_frame: bool
    channel: Optional[int | str]
    dlc: Optional[int]
    data: Optional[Base64Bytes]
    is_fd: bool
    is_rx: bool
    bitrate_switch: bool
    error_state_indicator: bool

    @classmethod
    def construct(cls, msg):
        return cls.model_construct(
            timestamp=msg.timestamp,
            arbitration_id=msg.arbitration_id,
            is_extended_id=msg.is_extended_id,
            is_remote_frame=msg.is_remote_frame,
            is_error_frame=msg.is_error_frame,
            channel=msg.channel,
            dlc=msg.dlc,
            data=msg.data,
            is_fd=msg.is_fd,
            is_rx=msg.is_rx,
            bitrate_switch=msg.bitrate_switch,
            error_state_indicator=msg.error_state_indicator,
        )


class IsoTpParams(BaseModel):
    """
    ISO-TP configuration parameters.
    """

    stmin: int = 0
    blocksize: int = 8
    tx_data_length: int = 8
    tx_data_min_length: int | None = None
    override_receiver_stmin: float | None = None
    rx_flowcontrol_timeout: int = 1000
    rx_consecutive_frame_timeout: int = 1000
    tx_padding: int | None = None
    wftmax: int = 0
    max_frame_size: int = 4095
    can_fd: bool = False
    bitrate_switch: bool = False
    default_target_address_type: isotp.TargetAddressType = isotp.TargetAddressType.Physical
    rate_limit_enable: bool = False
    rate_limit_max_bitrate: int = 10000000
    rate_limit_window_size: float = 0.2
    listen_mode: bool = False
    blocking_send: bool = False

    def apply(self, socket):
        socket.set_opts(
            optflag=None,
            frame_txtime=None,
            ext_address=None,
            txpad=self.tx_padding,
            rxpad=None,
            rx_ext_address=None,
            tx_stmin=None,
        )
        socket.set_fc_opts(bs=self.blocksize, stmin=self.stmin, wftmax=self.wftmax)
        socket.set_ll_opts(
            mtu=isotp.socket.LinkLayerProtocol.CAN_FD if self.can_fd else isotp.socket.LinkLayerProtocol.CAN,
            tx_dl=self.tx_data_length,
            tx_flags=None,
        )


class IsoTpMessage(BaseModel):
    """
    An ISO-TP CAN message.
    """

    data: Optional[Base64Bytes]


class IsoTpAddress(BaseModel):
    """
    An ISO-TP address set.
    """

    addressing_mode: AddressingMode
    txid: int | None
    rxid: int | None
    target_address: int | None
    source_address: int | None
    physical_id: int | None
    functional_id: int | None
    address_extension: int | None
    rx_only: bool
    tx_only: bool

    @classmethod
    def validate(cls, addr: isotp.Address):
        return cls(
            addressing_mode=addr._addressing_mode,
            txid=addr._txid,
            rxid=addr._rxid,
            target_address=addr._target_address,
            source_address=addr._source_address,
            physical_id=addr.physical_id  # ty: ignore[possibly-unbound-attribute]
            if hasattr(addr, "physical_id")
            else None,
            functional_id=addr.functional_id  # ty: ignore[possibly-unbound-attribute]
            if hasattr(addr, "functional_id")
            else None,
            address_extension=addr._address_extension,
            rx_only=addr._rx_only,
            tx_only=addr._tx_only,
        )

    def dump(self):
        return isotp.Address(
            addressing_mode=self.addressing_mode,
            txid=self.txid,
            rxid=self.rxid,
            target_address=self.target_address,
            source_address=self.source_address,
            physical_id=self.physical_id,
            functional_id=self.functional_id,
            address_extension=self.address_extension,
            rx_only=self.rx_only,
            tx_only=self.tx_only,
        )


class IsoTpAsymmetricAddress(BaseModel):
    """
    An asymmetric ISO-TP address.
    """

    tx_addr: IsoTpAddress
    rx_addr: IsoTpAddress

    @classmethod
    def validate(cls, addr: isotp.AsymmetricAddress):
        return cls(
            tx_addr=IsoTpAddress.validate(addr.tx_addr),
            rx_addr=IsoTpAddress.validate(addr.rx_addr),
        )

    def dump(self):
        return isotp.AsymmetricAddress(
            tx_addr=self.tx_addr.dump(),
            rx_addr=self.rx_addr.dump(),
        )



import os
from itertools import islice
from random import randbytes
from threading import Semaphore

import can
import isotp
import pytest

from jumpstarter_driver_can.common import IsoTpParams
from jumpstarter_driver_can.driver import Can, IsoTpPython, IsoTpSocket

from jumpstarter.common.utils import serve


def test_client_can_send_recv(request):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        client1.send(can.Message(data=b"hello"))

        assert client2.recv().data == b"hello"

        with pytest.raises(NotImplementedError):
            # not implemented on virtual bus
            client1.flush_tx_buffer()


def test_client_can_property(request):
    driver = Can(channel=request.node.name, interface="virtual")
    with serve(driver) as client, client:
        assert client.channel_info == driver.bus.channel_info
        assert client.state == driver.bus.state
        assert client.protocol == driver.bus.protocol

        with pytest.raises(NotImplementedError):
            # not implemented on virtual bus
            client.state = can.BusState.PASSIVE


def test_client_can_iterator(request):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        client1.send(can.Message(data=b"a"))
        client1.send(can.Message(data=b"b"))
        client1.send(can.Message(data=b"c"))

        assert [msg.data for msg in islice(client2, 3)] == [b"a", b"b", b"c"]


def test_client_can_filter(request):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        client2.set_filters([{"can_id": 0x1, "can_mask": 0x1, "extended": True}])

        client1.send(can.Message(arbitration_id=0x0, data=b"a"))
        client1.send(can.Message(arbitration_id=0x1, data=b"b"))
        client1.send(can.Message(arbitration_id=0x2, data=b"c"))

        assert client2.recv().data == b"b"


def test_client_can_notifier(request):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        sem = Semaphore(0)

        def listener(msg):
            assert msg.data == b"hello"
            sem.release()

        notifier = can.Notifier(client2, [listener])

        client1.send(can.Message(data=b"hello"))

        sem.acquire()
        notifier.stop()


def test_client_can_redirect(request):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        bus3 = can.interface.Bus(request.node.name + "_inner", interface="virtual")
        bus4 = can.interface.Bus(request.node.name + "_inner", interface="virtual")

        notifier = can.Notifier(client2, [can.RedirectReader(bus3)])

        client1.send(can.Message(data=b"hello"))

        assert bus4.recv().data == b"hello"

        notifier.stop()


@pytest.mark.parametrize(
    "msgs, expected",
    [
        ([can.Message(data=b"a"), can.Message(data=b"b")], [(1, b"a"), (1, b"b"), (1, b"a"), (1, b"b")]),
        (can.Message(data=b"a"), [(1, b"a"), (1, b"a"), (1, b"a"), (1, b"a")]),
    ],
)
def test_client_can_send_periodic_local(request, msgs, expected):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):

        def modifier_callback(msg):
            msg.arbitration_id = 1

        client1.send_periodic(
            msgs=msgs,
            period=0.1,
            duration=1,
            store_task=True,
            modifier_callback=modifier_callback,
        )

        assert [(msg.arbitration_id, msg.data) for msg in islice(client2, 4)] == expected


@pytest.mark.parametrize(
    "msgs, expected",
    [
        ([can.Message(data=b"a"), can.Message(data=b"b")], [(0, b"a"), (0, b"b"), (0, b"a"), (0, b"b")]),
        (can.Message(data=b"a"), [(0, b"a"), (0, b"a"), (0, b"a"), (0, b"a")]),
    ],
)
def test_client_can_send_periodic_remote(request, msgs, expected):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        task = client1.send_periodic(
            msgs=msgs,
            period=0.1,
            duration=1,
            autostart=False,
            store_task=True,
        )

        task.start()

        assert [(msg.arbitration_id, msg.data) for msg in islice(client2, 4)] == expected


@pytest.mark.parametrize("tx_data_length", [8, 64])
@pytest.mark.parametrize("blocking_send", [False, True])
def test_client_can_isotp(request, tx_data_length, blocking_send):
    with (
        serve(Can(channel=request.node.name, interface="virtual")) as client1,
        serve(Can(channel=request.node.name, interface="virtual")) as client2,
        client1,
        client2,
    ):
        notifier1 = can.Notifier(client1, [])
        notifier2 = can.Notifier(client2, [])

        params = IsoTpParams(
            max_frame_size=2048,
            tx_data_length=tx_data_length,
            blocking_send=blocking_send,
        )

        transport1 = isotp.NotifierBasedCanStack(
            client1,
            notifier1,
            address=isotp.Address(rxid=1, txid=2),
            params=params.model_dump(),
        )
        transport2 = isotp.NotifierBasedCanStack(
            client2,
            notifier2,
            address=isotp.Address(rxid=2, txid=1),
            params=params.model_dump(),
        )

        transport1.start()
        transport2.start()

        message = randbytes(params.max_frame_size)

        transport1.send(message, send_timeout=10)
        assert transport2.recv(block=True, timeout=10) == message

        transport1.stop()
        transport2.stop()

        notifier1.stop()
        notifier2.stop()


@pytest.mark.parametrize("blocking_send", [False, True])
@pytest.mark.parametrize(
    "addresses",
    [
        (None, None),
        (isotp.Address(rxid=3, txid=4), isotp.Address(rxid=4, txid=3)),
        (
            isotp.AsymmetricAddress(
                tx_addr=isotp.Address(rxid=5, txid=6, tx_only=True), rx_addr=isotp.Address(rxid=7, txid=8, rx_only=True)
            ),
            isotp.AsymmetricAddress(
                tx_addr=isotp.Address(rxid=8, txid=7, tx_only=True), rx_addr=isotp.Address(rxid=6, txid=5, rx_only=True)
            ),
        ),
    ],
)
def test_client_isotp(request, blocking_send, addresses):
    params = IsoTpParams(
        max_frame_size=2048,
        tx_data_length=64,
        blocking_send=blocking_send,
    )

    with (
        serve(
            IsoTpPython(
                channel=request.node.name, interface="virtual", address=isotp.Address(rxid=1, txid=2), params=params
            )
        ) as client1,
        serve(
            IsoTpPython(
                channel=request.node.name, interface="virtual", address=isotp.Address(rxid=2, txid=1), params=params
            )
        ) as client2,
    ):
        if addresses[0]:
            client1.set_address(addresses[0])
        if addresses[1]:
            client2.set_address(addresses[1])

        client1.start()
        client2.start()

        client1.available()
        client1.transmitting()

        message = randbytes(params.max_frame_size)

        client1.send(message, send_timeout=10)

        assert client2.recv(block=True, timeout=10) == message

        client1.stop_sending()
        client1.stop_receiving()

        client1.stop()
        client2.stop()


@pytest.mark.skipif(not os.path.exists("/sys/devices/virtual/net/vcan0"), reason="vcan0 not available")
@pytest.mark.parametrize("can_fd", [False, True])
def test_client_isotp_socket(request, can_fd):
    params = IsoTpParams(
        max_frame_size=2048,
        blocking_send=False,
        can_fd=can_fd,
    )

    with (
        serve(IsoTpSocket(channel="vcan0", address=isotp.Address(rxid=1, txid=2), params=params)) as client1,
        serve(IsoTpSocket(channel="vcan0", address=isotp.Address(rxid=2, txid=1), params=params)) as client2,
    ):
        client1.start()
        client2.start()

        message = randbytes(params.max_frame_size)

        client1.send(message, send_timeout=10)

        assert client2.recv(block=True, timeout=10) == message

        client1.stop()
        client2.stop()



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Callable, List, Optional, Sequence, Tuple
from uuid import UUID

import can
import isotp
from can.bus import _SelfRemovingCyclicTask
from pydantic import ConfigDict, validate_call

from .common import CanMessage, IsoTpAddress, IsoTpAsymmetricAddress, IsoTpMessage
from jumpstarter.client import DriverClient


@dataclass(kw_only=True)
class RemoteCyclicSendTask(can.broadcastmanager.CyclicSendTaskABC):
    client: CanClient
    uuid: UUID

    def start(self) -> None:
        self.client.call("_start_task", self.uuid)

    def stop(self) -> None:
        self.client.call("_stop_task", self.uuid)


@dataclass(kw_only=True)
class CanClient(DriverClient, can.BusABC):
    """
    A generic CAN client for sending/recieving traffic to/from an exported CAN bus.
    """

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        self._periodic_tasks: List[_SelfRemovingCyclicTask] = []
        self._filters = None
        self._is_shutdown: bool = False

    @property
    @validate_call(validate_return=True)
    def state(self) -> can.BusState:
        """
        The current state of the CAN bus.
        """
        return self.call("state")

    @state.setter
    @validate_call(validate_return=True)
    def state(self, value: can.BusState) -> None:
        """
        Set the state of the CAN bus.
        """
        self.call("state", value)

    @cached_property
    @validate_call(validate_return=True)
    def channel_info(self) -> str:
        """
        Get the CAN channel info.
        """
        return self.call("channel_info")

    @cached_property
    @validate_call(validate_return=True)
    def protocol(self) -> can.CanProtocol:
        """
        Get the CAN protocol supported by the bus.
        """
        return self.call("protocol")

    @validate_call(validate_return=True, config=ConfigDict(arbitrary_types_allowed=True))
    def _recv_internal(self, timeout: Optional[float]) -> Tuple[Optional[can.Message], bool]:
        msg, filtered = self.call("_recv_internal", timeout)
        if msg:
            return can.Message(**CanMessage.model_validate(msg).__dict__), filtered
        return None, filtered

    @validate_call(validate_return=True, config=ConfigDict(arbitrary_types_allowed=True))
    def send(self, msg: can.Message, timeout: Optional[float] = None) -> None:
        """
        Send an individual CAN message.
        """
        self.call("send", CanMessage.construct(msg), timeout)

    @validate_call(validate_return=True, config=ConfigDict(arbitrary_types_allowed=True))
    def _send_periodic_internal(
        self,
        msgs: Sequence[can.Message],
        period: float,
        duration: Optional[float] = None,
        autostart: bool = True,
        modifier_callback: Optional[Callable[[can.Message], None]] = None,
    ) -> can.broadcastmanager.CyclicSendTaskABC:
        if modifier_callback:
            return super()._send_periodic_internal(msgs, period, duration, autostart, modifier_callback)
        else:
            msgs = [CanMessage.construct(msg) for msg in msgs]
            return RemoteCyclicSendTask(
                client=self, uuid=self.call("_send_periodic_internal", msgs, period, duration, autostart)
            )

    # python-can bug
    # https://docs.pydantic.dev/2.8/errors/usage_errors/#typed-dict-version
    # @validate_call(validate_return=True)
    def _apply_filters(self, filters: Optional[can.typechecking.CanFilters]) -> None:
        self.call("_apply_filters", filters)

    @validate_call(validate_return=True)
    def flush_tx_buffer(self) -> None:
        """
        Flush the transmission buffer.
        """
        self.call("flush_tx_buffer")

    @validate_call(validate_return=True)
    def shutdown(self) -> None:
        """
        Shutdown the bus.
        """
        self.call("shutdown")
        super().shutdown()


@dataclass(kw_only=True)
class IsoTpClient(DriverClient):
    """
    An ISO-TP CAN client for sending/recieving ISO-TP frames to/from an exported CAN bus.
    """

    def start(self) -> None:
        """
        Start listening for messages.
        """
        self.call("start")

    def stop(self) -> None:
        """
        Stop listening for messages.
        """
        self.call("stop")

    def send(self, data: bytes, target_address_type: int | None = None, send_timeout: float | None = None) -> None:
        """
        Enqueue an ISO-TP frame to send over the CAN network.
        """
        return self.call("send", IsoTpMessage.model_construct(data=data), target_address_type, send_timeout)

    def recv(self, block: bool = False, timeout: float | None = None) -> bytes | None:
        """
        Dequeue an ISO-TP frame from the reception queue if available.
        """
        return IsoTpMessage.model_validate(self.call("recv", block, timeout)).data

    def available(self) -> bool:
        """
        Returns `True` if an ISO-TP frame is awaiting in the reception queue, `False` otherwise.
        """
        return self.call("available")

    def transmitting(self) -> bool:
        """
        Returns `True` if an ISO-TP frame is being transmitted, `False` otherwise.
        """
        return self.call("transmitting")

    def set_address(self, address: isotp.Address | isotp.AsymmetricAddress) -> None:
        """
        Sets the layer address. Can be set after initialization if needed.
        May cause a timeout if called while a transmission is active.
        """
        match address:
            case isotp.Address():
                return self.call("set_address", IsoTpAddress.validate(address))
            case isotp.AsymmetricAddress():
                return self.call("set_address", IsoTpAsymmetricAddress.validate(address))
            case _:
                raise ValueError("address not isotp.Address | isotp.AsymmetricAddress")

    def stop_sending(self) -> None:
        """
        Stop sending messages.
        """
        self.call("stop_sending")

    def stop_receiving(self) -> None:
        """
        Stop receiving messages.
        """
        self.call("stop_receiving")



from dataclasses import field
from typing import Callable, Optional, Sequence, Tuple, Union
from uuid import UUID, uuid4

import can
import isotp
from pydantic import ConfigDict, validate_call
from pydantic.dataclasses import dataclass

from .common import CanMessage, IsoTpAddress, IsoTpAsymmetricAddress, IsoTpMessage, IsoTpParams
from jumpstarter.driver import Driver, export


@dataclass(kw_only=True, config=ConfigDict(arbitrary_types_allowed=True))
class Can(Driver):
    """
    A generic CAN bus driver.

    Available on any platform, supports many different CAN
    interfaces through the `python-can` library.
    """

    """
    The CAN channel to connect to.
    """
    channel: str | int | None

    """
    The CAN interface to use.
    """
    interface: str | None

    """
    The CAN bus instance used for communication.
    """
    bus: can.BusABC = field(init=False)

    """
    A dict of cyclic send tasks to run.
    """
    __tasks: dict[UUID, can.broadcastmanager.CyclicSendTaskABC] = field(init=False, default_factory=dict)

    @classmethod
    def client(cls) -> str:
        return "jumpstarter_driver_can.client.CanClient"

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        self.bus = can.Bus(channel=self.channel, interface=self.interface)

    @export
    @validate_call(validate_return=True)
    def _recv_internal(self, timeout: Optional[float]) -> Tuple[Optional[CanMessage], bool]:
        msg, filtered = self.bus._recv_internal(timeout)
        if msg:
            return CanMessage.construct(msg), filtered
        return None, filtered

    @export
    @validate_call(validate_return=True)
    def send(self, msg: CanMessage, timeout: float | None = None):
        """
        Send an individual CAN message.
        """
        self.bus.send(can.Message(**msg.__dict__), timeout)

    @export
    @validate_call(validate_return=True, config=ConfigDict(arbitrary_types_allowed=True))
    def _send_periodic_internal(
        self,
        msgs: Union[Sequence[CanMessage], CanMessage],
        period: float,
        duration: Optional[float] = None,
        autostart: bool = True,
        modifier_callback: Optional[Callable[[can.Message], None]] = None,
    ) -> UUID:
        assert modifier_callback is None
        task = self.bus._send_periodic_internal(msgs, period, duration, autostart, modifier_callback)
        uuid = uuid4()
        self.__tasks[uuid] = task
        return uuid

    @export
    @validate_call(validate_return=True)
    def _start_task(self, uuid: UUID) -> None:
        self.__tasks[uuid].start()

    @export
    @validate_call(validate_return=True)
    def _stop_task(self, uuid: UUID) -> None:
        self.__tasks.pop(uuid).stop()

    @export
    @validate_call(validate_return=True)
    def state(self, value: can.BusState | None = None) -> can.BusState | None:
        """
        The current state of the CAN bus.
        """
        if value:
            self.bus.state = value
            return None
        else:
            return self.bus.state

    @export
    @validate_call(validate_return=True)
    def protocol(self) -> can.CanProtocol:
        """
        Get the CAN protocol supported by the bus.
        """
        return self.bus.protocol

    @export
    @validate_call(validate_return=True)
    def channel_info(self) -> str:
        """
        Get the CAN channel info.
        """
        return self.bus.channel_info

    @export
    # python-can bug
    # https://docs.pydantic.dev/2.8/errors/usage_errors/#typed-dict-version
    # @validate_call(validate_return=True)
    def _apply_filters(self, filters: Optional[can.typechecking.CanFilters]) -> None:
        self.bus._apply_filters(filters)

    @export
    @validate_call(validate_return=True)
    def flush_tx_buffer(self) -> None:
        """
        Flush the transmission buffer.
        """
        self.bus.flush_tx_buffer()

    @export
    @validate_call(validate_return=True)
    def shutdown(self) -> None:
        """
        Shutdown the bus.
        """
        self.bus.shutdown()


@dataclass(kw_only=True, config=ConfigDict(arbitrary_types_allowed=True))
class IsoTpPython(Driver):
    """
    Pure python ISO-TP socket driver

    Available on any platform, moderate performance and reliability,
    wide support for non-standard hardware interfaces
    """

    """
    The CAN channel to connect to.
    """
    channel: str | int | None

    """
    The CAN interface to use.
    """
    interface: str | None

    """
    The ISO-TP addressing to use.
    """
    address: isotp.Address

    """
    The ISO-TP parameters.
    """
    params: IsoTpParams = field(default_factory=IsoTpParams)

    """
    The read timeout for the bus.
    """
    read_timeout: float = 0.05

    """
    The CAN bus instance used for communication.
    """
    bus: can.BusABC = field(init=False)

    """
    The CAN bus notifier instance used to receive messages.
    """
    notifier: can.Notifier = field(init=False)

    """
    The ISO-TP CAN layer configured to use `python-can`.
    """
    stack: isotp.NotifierBasedCanStack = field(init=False)

    @classmethod
    def client(cls) -> str:
        return "jumpstarter_driver_can.client.IsoTpClient"

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        self.bus = can.Bus(channel=self.channel, interface=self.interface)
        self.notifier = can.Notifier(self.bus, [])
        self.stack = isotp.NotifierBasedCanStack(
            self.bus,
            self.notifier,
            address=self.address,
            params=self.params.model_dump(),
            read_timeout=self.read_timeout,
        )

    @export
    @validate_call(validate_return=True)
    def start(self) -> None:
        """
        Start listening for messages.
        """
        self.stack.start()

    @export
    @validate_call(validate_return=True)
    def stop(self) -> None:
        """
        Stop listening for messages.
        """
        self.stack.stop()

    @export
    @validate_call(validate_return=True)
    def send(
        self, msg: IsoTpMessage, target_address_type: int | None = None, send_timeout: float | None = None
    ) -> None:
        """
        Enqueue an ISO-TP frame to send over the CAN network.
        """
        return self.stack.send(msg.data, target_address_type, send_timeout)

    @export
    @validate_call(validate_return=True)
    def recv(self, block: bool = False, timeout: float | None = None) -> IsoTpMessage:
        """
        Dequeue an ISO-TP frame from the reception queue if available.
        """
        return IsoTpMessage.model_construct(data=self.stack.recv(block, timeout))

    @export
    @validate_call(validate_return=True)
    def available(self) -> bool:
        """
        Returns `True` if an ISO-TP frame is awaiting in the reception queue, `False` otherwise.
        """
        return self.stack.available()

    @export
    @validate_call(validate_return=True)
    def transmitting(self) -> bool:
        """
        Returns `True` if an ISO-TP frame is being transmitted, `False` otherwise.
        """
        return self.stack.transmitting()

    @export
    @validate_call(validate_return=True)
    def set_address(self, address: IsoTpAddress | IsoTpAsymmetricAddress) -> None:
        """
        Sets the layer address. Can be set after initialization if needed.
        May cause a timeout if called while a transmission is active.
        """
        self.stack.set_address(address.dump())

    @export
    @validate_call(validate_return=True)
    def stop_sending(self) -> None:
        """
        Stop sending messages.
        """
        self.stack.stop_sending()

    @export
    @validate_call(validate_return=True)
    def stop_receiving(self) -> None:
        """
        Stop receiving messages.
        """
        self.stack.stop_receiving()


@dataclass(kw_only=True, config=ConfigDict(arbitrary_types_allowed=True))
class IsoTpSocket(Driver):
    """
    Linux kernel ISO-TP socket driver

    Available since kernel 5.10, good performance and reliability,
    only supports standard hardware interfaces
    """

    """
    The CAN channel to connect to.
    """
    channel: str

    """
    The ISO-TP addressing to use.
    """
    address: isotp.Address
    params: IsoTpParams = field(default_factory=IsoTpParams)

    sock: isotp.socket | None = field(init=False, default=None)

    @classmethod
    def client(cls) -> str:
        return "jumpstarter_driver_can.client.IsoTpClient"

    @export
    @validate_call(validate_return=True)
    def start(self) -> None:
        """
        Start listening for messages.
        """
        if self.sock:
            raise ValueError("socket already started")
        self.sock = isotp.socket()
        self.params.apply(self.sock)
        self.sock.bind(self.channel, self.address)

    @export
    @validate_call(validate_return=True)
    def stop(self) -> None:
        """
        Stop listening for messages.
        """
        if not self.sock:
            raise ValueError("socket not started")
        self.sock.close()
        self.sock = None

    @export
    @validate_call(validate_return=True)
    def send(
        self, msg: IsoTpMessage, target_address_type: int | None = None, send_timeout: float | None = None
    ) -> None:
        """
        Enqueue an ISO-TP frame to send over the CAN network.
        """
        if not self.sock:
            raise ValueError("socket not started")
        self.sock.send(msg.data)

    @export
    @validate_call(validate_return=True)
    def recv(self, block: bool = False, timeout: float | None = None) -> IsoTpMessage:
        """
        Dequeue an ISO-TP frame from the reception queue if available.
        """
        if not self.sock:
            raise ValueError("socket not started")
        return IsoTpMessage.model_construct(data=self.sock.recv())

    @export
    @validate_call(validate_return=True)
    def available(self) -> bool:
        """
        Not supported by the socket ISO-TP driver.
        """
        raise NotImplementedError

    @export
    @validate_call(validate_return=True)
    def transmitting(self) -> bool:
        """
        Not supported by the socket ISO-TP driver.
        """
        raise NotImplementedError

    @export
    @validate_call(validate_return=True)
    def set_address(self, address: IsoTpAddress | IsoTpAsymmetricAddress) -> None:
        """
        Not supported by the socket ISO-TP driver.
        """
        raise NotImplementedError

    @export
    @validate_call(validate_return=True)
    def stop_sending(self) -> None:
        """
        Not supported by the socket ISO-TP driver.
        """
        raise NotImplementedError

    @export
    @validate_call(validate_return=True)
    def stop_receiving(self) -> None:
        """
        Not supported by the socket ISO-TP driver.
        """
        raise NotImplementedError
